"""Use the GraphQL api to grab issues/PRs that match a query."""
import datetime
import dateutil
import pytz
import requests
import urllib
from pathlib import Path

from .graphql import GitHubGraphQlQuery
from .cache import _cache_data
import pandas as pd
import numpy as np


# The tags and description to use in creating subsets of PRs
TAGS_METADATA_BASE = {
    "new": {
        "tags": ["feature", "new"],
        "pre": ["NEW", "FEAT", "FEATURE"],
        "description": "New features added",
    },
    "enhancement": {
        "tags": ["enhancement", "enhancements"],
        "pre": ["NEW", "ENH", "ENHANCEMENT", "IMPROVE"],
        "description": "Enhancements made",
    },
    "bug": {
        "tags": ["bug", "bugfix", "bugs"],
        "pre": ["FIX", "BUG"],
        "description": "Bugs fixed",
    },
    "maintenance": {
        "tags": ["maintenance", "maint"],
        "pre": ["MAINT", "MNT"],
        "description": "Maintenance and upkeep improvements",
    },
    "documentation": {
        "tags": ["documentation", "docs", "doc"],
        "pre": ["DOC", "DOCS"],
        "description": "Documentation improvements",
    },
    "api_change": {
        "tags": ["api-change", "apichange"],
        "pre": ["BREAK", "BREAKING", "BRK", "UPGRADE"],
        "description": "API and Breaking Changes",
    },
    "deprecate": {
        "tags": ["deprecation", "deprecate"],
        "pre": ["DEPRECATE", "DEPRECATION", "DEP"],
        "description": "Deprecated features",
    },
}


def get_activity(
    target, since, until=None, repo=None, kind=None, auth=None, cache=None
):
    """Return issues/PRs within a date window.

    Parameters
    ----------
    target : string
        The GitHub organization/repo for which you want to grab recent issues/PRs.
        Can either be *just* and organization (e.g., `jupyter`) or a combination
        organization and repo (e.g., `jupyter/notebook`). If the former, all
        repositories for that org will be used. If the latter, only the specified
        repository will be used.
    since : string | None
        Return issues/PRs with activity since this date or git reference. Can be
        any string that is parsed with dateutil.parser.parse.
    until : string | None
        Return issues/PRs with activity until this date or git reference. Can be
        any string that is parsed with dateutil.parser.parse. If none, today's
        date will be used.
    kind : ["issue", "pr"] | None
        Return only issues or PRs. If None, both will be returned.
    auth : string | None
        An authentication token for GitHub. If None, then the environment
        variable `GITHUB_ACCESS_TOKEN` will be tried.
    cache : bool | str | None
        Whether to cache the returned results. If None, no caching is
        performed. If True, the cache is located at
        ~/github_activity_data. It is organized as orgname/reponame folders
        with CSV files inside that contain the latest data. If a string it
        is treated as the path to a cache folder.

    Returns
    -------
    query_data : pandas DataFrame
        A munged collection of data returned from your query. This
        will be a combination of issues and PRs.
    """

    org, repo = _parse_target(target)
    if repo:
        # We have org/repo
        search_query = f"repo:{org}/{repo}"
    else:
        # We have just org
        search_query = f"user:{org}"

    # Figure out dates for our query
    since_dt, since_is_git_ref = _get_datetime_and_type(org, repo, since)
    until_dt, until_is_git_ref = _get_datetime_and_type(org, repo, until)
    since_dt_str = f"{since_dt:%Y-%m-%dT%H:%M:%SZ}"
    until_dt_str = f"{until_dt:%Y-%m-%dT%H:%M:%SZ}"

    if kind:
        allowed_kinds = ["issue", "pr"]
        if kind not in allowed_kinds:
            raise ValueError(f"Kind must be one of {allowed_kinds}")
        search_query += f" type:{kind}"

    # Query for both opened and closed issues/PRs in this window
    print(f"Running search query:\n{search_query}\n\n")
    query_data = []
    for activity_type in ["created", "closed"]:
        ii_search_query = (
            search_query + f" {activity_type}:{since_dt_str}..{until_dt_str}"
        )
        qu = GitHubGraphQlQuery(ii_search_query, auth=auth)
        qu.request()
        query_data.append(qu.data)

    query_data = (
        pd.concat(query_data).drop_duplicates(subset=["id"]).reset_index(drop=True)
    )
    query_data.since_dt = since_dt
    query_data.until_dt = until_dt
    query_data.since_dt_str = since_dt_str
    query_data.until_dt_str = until_dt_str
    query_data.since_is_git_ref = since_is_git_ref
    query_data.until_is_git_ref = until_is_git_ref

    if cache:
        _cache_data(query_data, cache)
    return query_data


def generate_activity_md(
    target,
    since=None,
    until=None,
    kind=None,
    auth=None,
    tags=None,
    include_issues=False,
    include_opened=False,
    strip_brackets=False,
):
    """Generate a markdown changelog of GitHub activity within a date window.

    Parameters
    ----------
    target : string
        The GitHub organization/repo for which you want to grab recent issues/PRs.
        Can either be *just* and organization (e.g., `jupyter`) or a combination
        organization and repo (e.g., `jupyter/notebook`). If the former, all
        repositories for that org will be used. If the latter, only the specified
        repository will be used. Can also be a URL to a GitHub org or repo.
    since : string | None
        Return issues/PRs with activity since this date or git reference. Can be
        any string that is parsed with dateutil.parser.parse. If None, the date
        of the latest release will be used.
    until : string | None
        Return issues/PRs with activity until this date or git reference. Can be
        any string that is parsed with dateutil.parser.parse. If none, today's
        date will be used.
    kind : ["issue", "pr"] | None
        Return only issues or PRs. If None, both will be returned.
    auth : string | None
        An authentication token for GitHub. If None, then the environment
        variable `GITHUB_ACCESS_TOKEN` will be tried.
    tags : list of strings | None
        A list of the tags to use in generating subsets of PRs for the markdown report.
        Must be one of:

            ['enhancement', 'bugs', 'maintenance', 'documentation', 'api_change']

        If None, all of the above tags will be used.
    include_issues : bool
        Include Issues in the markdown output. Default is False.
    include_opened : bool
        Include a list of opened items in the markdown output. Default is False.
    strip_brackets : bool
        If True, strip any text between brackets at the beginning of the issue/PR title.
        E.g., [MRG], [DOC], etc.

    Returns
    -------
    query_data : pandas DataFrame
        A munged collection of data returned from your query. This
        will be a combination of issues and PRs.
    """
    org, repo = _parse_target(target)

    # If no since parameter is given, find the name of the latest release
    if since is None:
        since = _get_latest_tag(org, repo)

    # Grab the data according to our query
    data = get_activity(
        target, since=since, until=until, kind=kind, auth=auth, cache=False
    )
    if data.empty:
        return

    # Collect authors of comments on issues/prs that they didn't open for our attribution list
    comment_response_cutoff = 6  # Comments on a single issue
    comment_others_cutoff = 2  # Comments on issues somebody else has authored
    comment_helpers = []
    all_contributors = []
    for _, iitems in data.iterrows():
        item_commentors = []
        for icomment in iitems["comments"]["edges"]:
            comment_author = icomment["node"]["author"]
            if not comment_author:
                # This happens if the GitHub user has been deleted
                # ref: https://github.com/jupyterhub/oauthenticator/pull/224#issuecomment-453211986
                continue

            comment_author = comment_author["login"]

            # Add to list of commentors on items they didn't author
            if comment_author != iitems["author"]:
                comment_helpers.append(comment_author)

            # Add to list of commentors for this item so we can see how many times they commented
            item_commentors.append(comment_author)

        # Count any commentors that had enough comments on the issue to be a contributor
        item_commentors_counts = pd.value_counts(item_commentors)
        item_commentors_counts = item_commentors_counts[
            item_commentors_counts >= comment_response_cutoff
        ].index.tolist()
        for person in item_commentors_counts:
            all_contributors.append(person)

    comment_contributor_counts = pd.value_counts(comment_helpers)
    all_contributors += comment_contributor_counts[
        comment_contributor_counts >= comment_others_cutoff
    ].index.tolist()

    # Clean up the data a bit
    data["labels"] = data["labels"].map(
        lambda a: [edge["node"]["name"] for edge in a["edges"]]
    )
    data["kind"] = data["url"].map(lambda a: "issue" if "issues/" in a else "pr")

    # Separate into closed and opened
    until_dt_str = data.until_dt_str
    since_dt_str = data.since_dt_str
    closed = data.query("closedAt >= @since_dt_str and closedAt <= @until_dt_str")
    opened = data.query("createdAt >= @since_dt_str and createdAt <= @until_dt_str")

    # Separate into PRs and issues
    closed_prs = closed.query("kind == 'pr'")
    closed_issues = closed.query("kind == 'issue'")
    opened_prs = opened.query("kind == 'pr'")
    opened_issues = opened.query("kind == 'issue'")

    # Remove the PRs/Issues that from "opened" if they were also closed
    mask_open_and_close_pr = opened_prs["id"].map(
        lambda iid: iid in closed_prs["id"].values
    )
    mask_open_and_close_issue = opened_issues["id"].map(
        lambda iid: iid in closed_issues["id"].values
    )
    opened_prs = opened_prs.loc[~mask_open_and_close_pr]
    opened_issues = opened_issues.loc[~mask_open_and_close_issue]

    # Now remove the *closed* PRs (not merged) for our output list
    closed_prs = closed_prs.query("state != 'CLOSED'")

    # Add any author of a merged PR to our contributors list
    all_contributors += closed_prs["author"].unique().tolist()

    # Define categories for a few labels
    if tags is None:
        tags = TAGS_METADATA_BASE.keys()
    if not all(tag in TAGS_METADATA_BASE for tag in tags):
        raise ValueError(
            "You provided an unsupported tag. Tags must be "
            f"one or more of {TAGS_METADATA_BASE.keys()}, You provided:\n"
            f"{tags}"
        )
    tags_metadata = {key: val for key, val in TAGS_METADATA_BASE.items() if key in tags}

    # Initialize our tags with empty metadata
    for key, vals in tags_metadata.items():
        vals.update(
            {"mask": None, "md": [], "data": None,}
        )

    # Separate out items by their tag types
    for kind, kindmeta in tags_metadata.items():
        # First find the PRs based on tag
        mask = closed_prs["labels"].map(
            lambda a: any(ii in jj for ii in kindmeta["tags"] for jj in a)
        )
        # Now find PRs based on prefix
        mask_pre = closed_prs["title"].map(
            lambda title: any(f"{ipre}:" in title for ipre in kindmeta["pre"])
        )
        mask = mask | mask_pre

        kindmeta["data"] = closed_prs.loc[mask]
        kindmeta["mask"] = mask

    # All remaining PRs w/o a label go here
    all_masks = np.array(
        [~kindinfo["mask"].values for _, kindinfo in tags_metadata.items()]
    )
    mask_others = all_masks.all(0)
    others = closed_prs.loc[mask_others]
    other_description = (
        "Other merged PRs" if len(others) != len(closed_prs) else "Merged PRs"
    )

    # Add some optional kinds of PRs / issues
    tags_metadata.update(
        dict(others={"description": other_description, "md": [], "data": others})
    )
    if include_issues:
        tags_metadata.update(
            dict(
                closed_issues={
                    "description": "Closed issues",
                    "md": [],
                    "data": closed_issues,
                }
            )
        )
        if include_opened:
            tags_metadata.update(
                dict(
                    opened_issues={
                        "description": "Opened issues",
                        "md": [],
                        "data": opened_issues,
                    }
                )
            )
    if include_opened:
        tags_metadata.update(
            dict(opened_prs={"description": "Opened PRs", "md": [], "data": opened_prs})
        )

    # Generate the markdown
    prs = tags_metadata

    for kind, items in prs.items():
        n_orgs = len(items["data"]["org"].unique())
        for org, idata in items["data"].groupby("org"):
            if n_orgs > 1:
                items["md"].append(f"## {org}")
                items["md"].append("")

            for irow, irowdata in items["data"].iterrows():
                author = irowdata["author"]
                ititle = irowdata["title"]
                if strip_brackets and ititle.strip().startswith("[") and "]" in ititle:
                    ititle = ititle.split("]", 1)[-1].strip()
                this_md = f"* {ititle} [#{irowdata['number']}]({irowdata['url']}) ([@{author}](https://github.com/{author}))"
                items["md"].append(this_md)

    # Get functional GitHub references: any git reference or master@{YY-mm-dd}
    if not data.since_is_git_ref:
        since = f"master@{{{data.since_dt:%Y-%m-%d}}}"
        closest_date_start = closed_prs.loc[
            abs(
                pd.to_datetime(closed_prs["closedAt"], utc=True)
                - pd.to_datetime(data.since_dt, utc=True)
            ).idxmin()
        ]
        since_ref = closest_date_start["mergeCommit"]["oid"]
    else:
        since_ref = since

    if not data.until_is_git_ref:
        until = f"master@{{{data.until_dt:%Y-%m-%d}}}"
        closest_date_stop = closed_prs.loc[
            abs(
                pd.to_datetime(closed_prs["closedAt"], utc=True)
                - pd.to_datetime(data.until_dt, utc=True)
            ).idxmin()
        ]
        until_ref = closest_date_stop["mergeCommit"]["oid"]
    else:
        until_ref = until
    # SHAs for our dates to build the GitHub diff URL
    changelog_url = f"https://github.com/{org}/{repo}/compare/{since_ref}...{until_ref}"

    # Build the Markdown
    md = [f"# {since}...{until}", f"([full changelog]({changelog_url}))", ""]
    for kind, info in prs.items():
        if len(info["md"]) > 0:
            md += [""]
            md.append(f"## {info['description']}")
            md += info["md"]

    # Add a list of author contributions
    all_contributors = sorted(set(all_contributors), key=lambda a: str(a).lower())
    all_contributor_links = []
    for iauthor in all_contributors:
        author_url = f"https://github.com/search?q=repo%3A{org}%2F{repo}+involves%3A{iauthor}+updated%3A{data.since_dt:%Y-%m-%d}..{data.until_dt:%Y-%m-%d}&type=Issues"
        all_contributor_links.append(f"[@{iauthor}]({author_url})")
    contributor_md = " | ".join(all_contributor_links)
    gh_contributors_link = f"https://github.com/{org}/{repo}/graphs/contributors?from={data.since_dt:%Y-%m-%d}&to={data.until_dt:%Y-%m-%d}&type=c"
    md += [""]
    md += ["## Contributors to this release"]
    md += [f"([GitHub contributors page for this release]({gh_contributors_link}))"]
    md += [""]
    md += [contributor_md]
    md = "\n".join(md)
    return md


def extract_comments(comments):
    """Extract the comments returned from GraphQL Issues or PullRequests.

    Parameters
    ----------
    comments: pandas Series
        The comments column in the DataFrame returned by `get_activity`.

    Returns
    -------
    comments: pandas DataFrame
        Each comment with metadata for the query given.
    """
    list_of_comments = [ii["edges"] for ii in comments]
    has_comments = any(jj.get("node") for ii in list_of_comments for jj in ii)

    # If we have no comments, just return None
    if not has_comments:
        return None

    comments = [jj.get("node") for ii in list_of_comments for jj in ii]
    comments = pd.DataFrame(comments)
    comments["author"] = comments["author"].map(
        lambda a: a["login"] if a is not None else a
    )

    # Parse some data about the comments
    url_parts = [ii.split("/") for ii in comments["url"].values]
    url_parts = np.array([(ii[3], ii[4], ii[6]) for ii in url_parts])
    orgs, repos, url_parts = url_parts.T

    issue_id = [ii.split("#")[0] for ii in url_parts]
    comment_id = [ii.split("-")[-1] for ii in url_parts]

    # Assign new variables
    comments["org"] = orgs
    comments["repo"] = repos
    comments["issue_id"] = issue_id
    comments["id"] = comment_id
    return comments


def _parse_target(target):
    if target.startswith("http"):
        target = target.split("github.com/")[-1]
    parts = target.split("/")
    if len(parts) == 2:
        org, repo = parts
    elif len(parts) == 1:
        (org,) = parts
        repo = None
    else:
        raise ValueError(
            f"Invalid target. Target should be of the form org/repo or a GitHub URL. Got {target}"
        )
    return org, repo


def _get_datetime_and_type(org, repo, datetime_or_git_ref):
    """Return a datetime object and bool."""

    is_git_ref = False
    if datetime_or_git_ref is None:
        dt = datetime.datetime.now().astimezone(pytz.utc)
    else:
        try:
            dt = dateutil.parser.parse(datetime_or_git_ref)
        except ValueError as e:
            if repo:
                dt = _get_datetime_from_git_ref(org, repo, datetime_or_git_ref)
                is_git_ref = True
            else:
                raise e

    return dt, is_git_ref


def _get_datetime_from_git_ref(org, repo, ref):
    """Return a datetime from a git reference."""

    response = requests.get(f"https://api.github.com/repos/{org}/{repo}/commits/{ref}")
    response.raise_for_status()
    return dateutil.parser.parse(response.json()["commit"]["committer"]["date"])


def _get_latest_tag(org, repo):
    """Return the latest tag name for a given repository."""
    response = requests.get(f"https://api.github.com/repos/{org}/{repo}/git/refs/tags")
    response.raise_for_status()
    tags = response.json()
    latest_tag = list(tags)[-1]
    return latest_tag["ref"].split("/tags/")[-1]



import argparse
import os
import sys

from .github_activity import generate_activity_md

DESCRIPTION = "Generate a markdown changelog of GitHub activity within a date window."
parser = argparse.ArgumentParser(description=DESCRIPTION)
parser.add_argument(
    "target",
    help="""The GitHub organization/repo for which you want to grab recent issues/PRs.
    Can either be *just* an organization (e.g., `jupyter`), or a combination
    organization and repo (e.g., `jupyter/notebook`). If the former, all
    repositories for that org will be used. If the latter, only the specified
    repository will be used. Can also be a GitHub URL to an organization or repo.""",
)
parser.add_argument(
    "-s",
    "--since",
    nargs="?",
    default=None,
    help="""Return issues/PRs with activity since this date or git reference. Can be
    any string that is parsed with dateutil.parser.parse.""",
)
parser.add_argument(
    "-u",
    "--until",
    nargs="?",
    default=None,
    help="""Return issues/PRs with activity until this date or git reference. Can be
    any string that is parsed with dateutil.parser.parse. If none, today's
    date will be used.""",
)
parser.add_argument(
    "-o", "--output", default=None, help="""Write the markdown to a file if desired."""
)
parser.add_argument(
    "--kind",
    default=None,
    help="""Return only issues or PRs. If None, both will be returned.""",
)
parser.add_argument(
    "--auth",
    default=None,
    help=(
        "An authentication token for GitHub. If None, then the environment "
        "variable `GITHUB_ACCESS_TOKEN` will be tried."
    ),
)
parser.add_argument(
    "--tags",
    default=None,
    help=(
        "A list of the tags to use in generating subsets of PRs for the "
        "markdown report. Must be one of:"
        ""
        "   ['enhancement', 'bugs', 'maintenance', 'documentation', 'api_change']"
        ""
        "If None, all of the above tags will be used."
    ),
)
parser.add_argument(
    "--include-issues",
    default=False,
    action="store_true",
    help="Include Issues in the markdown output",
)
parser.add_argument(
    "--include-opened",
    default=False,
    action="store_true",
    help="Include a list of opened items in the markdown output",
)
parser.add_argument(
    "--strip-brackets",
    default=False,
    action="store_true",
    help=(
        "If True, strip any text between brackets at the beginning of the issue/PR title. "
        "E.g., [MRG], [DOC], etc."
    ),
)


def main():
    args = parser.parse_args(sys.argv[1:])
    tags = args.tags.split(",") if args.tags is not None else args.tags
    md = generate_activity_md(
        args.target,
        since=args.since,
        until=args.until,
        kind=args.kind,
        auth=args.auth,
        tags=tags,
        include_issues=bool(args.include_issues),
        include_opened=bool(args.include_opened),
        strip_brackets=bool(args.strip_brackets),
    )
    if not md:
        return

    if args.output:
        output = os.path.abspath(args.output)
        output_dir = os.path.dirname(output)
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        with open(args.output, "w") as ff:
            ff.write(md)
        print(f"Finished writing markdown to: {args.output}")
    else:
        print(md)


if __name__ == "__main__":
    main()



import os
import requests
from datetime import timedelta

import numpy as np
import pandas as pd
from IPython.display import display
from ipywidgets import widgets

comments_query = """\
        comments(last: 100) {
          edges {
            node {
              authorAssociation
              createdAt
              updatedAt
              url
              author {
                login
              }
            }
          }
        }
"""

base_elements = """\
        state
        id
        title
        url
        createdAt
        updatedAt
        closedAt
        labels(first: 10) {
            edges {
                node {
                    name
                }
            }
        }
        number
        authorAssociation
        author {
          login
        }
        reactions(content: THUMBS_UP) {
          totalCount
        }
"""

gql_template = """\
{{
  search({query}) {{
    issueCount
    pageInfo {{
        endCursor
        hasNextPage
    }}
    nodes {{
      ... on PullRequest {{
        {base_elements}
        mergedBy {{
          login
        }}
        mergeCommit {{
          oid
        }}
        {comments}
      }}
      ... on Issue {{
        {base_elements}
        {comments}
      }}
    }}
  }}
}}
"""

# Define our query object that we'll re-use for github search
class GitHubGraphQlQuery:
    def __init__(self, query, display_progress=True, auth=None):
        """Run a GitHub GraphQL query and return the issue/PR data from it.

        Parameters
        ----------
        query : string
          The GitHub search query to run. This is similar to whatever you'd use
          to search on GitHub.com.
        display_progress : bool
          Whether to display a progress bar as data is fetched.
        auth : string | None
          An authentication token for GitHub. If None, then the environment
          variable `GITHUB_ACCESS_TOKEN` will be tried.
        """
        self.query = query

        # Authentication
        auth = auth or os.environ.get("GITHUB_ACCESS_TOKEN")
        if not auth:
            raise ValueError(
                "Either the environment variable GITHUB_ACCESS_TOKEN or the "
                "--auth flag or must be used to pass a Personal Access Token "
                "needed by the GitHub API. You can generate a token at "
                "https://github.com/settings/tokens/new. Note that while "
                "working with a public repository, you don’t need to set any "
                "scopes on the token you create."
            )
        self.headers = {"Authorization": "Bearer %s" % auth}

        self.gql_template = gql_template
        self.display_progress = display_progress

    def request(self, n_pages=100, n_per_page=50):
        """Make a request to the GitHub GraphQL API.

        This generates an attribute `self.data` with a pandas
        DataFrame of the issue / PR activity corresponding to
        the query you ran.
        """

        # NOTE: This main search query has a type, but the query string also has a type.
        # ref ("search"): https://developer.github.com/v4/query/#connections
        # Collect paginated issues
        self.issues_and_or_prs = []
        for ii in range(n_pages):
            github_search_query = [
                "first: %s" % n_per_page,
                'query: "%s"' % self.query,
                "type: ISSUE",
            ]
            if ii != 0:
                github_search_query.append('after: "%s"' % pageInfo["endCursor"])

            ii_gql_query = self.gql_template.format(
                query=", ".join(github_search_query),
                comments=comments_query,
                base_elements=base_elements,
            )
            ii_request = requests.post(
                "https://api.github.com/graphql",
                json={"query": ii_gql_query},
                headers=self.headers,
            )
            if ii_request.status_code != 200:
                raise Exception(
                    "Query failed to run by returning code of {}. {}".format(
                        ii_request.status_code, ii_gql_query
                    )
                )
            if "errors" in ii_request.json().keys():
                raise Exception(
                    "Query failed to run with error {}. {}".format(
                        ii_request.json()["errors"], ii_gql_query
                    )
                )
            self.last_request = ii_request

            # Parse the response for this pagination
            json = ii_request.json()["data"]["search"]
            if ii == 0:
                if json["issueCount"] == 0:
                    print("Found no entries for query.")
                    self.data = pd.DataFrame()
                    return

                n_pages = int(np.ceil(json["issueCount"] / n_per_page))
                print(
                    "Found {} items, which will take {} pages".format(
                        json["issueCount"], n_pages
                    )
                )
                prog = widgets.IntProgress(
                    value=0,
                    min=0,
                    max=n_pages,
                    description="Downloading:",
                    bar_style="",
                )
                if n_pages > 1 and self.display_progress:
                    display(prog)

            # Add the JSON to the raw data list
            self.issues_and_or_prs.extend(json["nodes"])
            pageInfo = json["pageInfo"]
            self.last_query = ii_gql_query

            # Update progress and should we stop?
            prog.value += 1
            if pageInfo["hasNextPage"] is False:
                prog.bar_style = "success"
                break

        # Create a dataframe of the issues and/or PRs
        self.data = pd.DataFrame(self.issues_and_or_prs)

        # Add some extra fields
        self.data["author"] = self.data["author"].map(
            lambda a: a["login"] if a is not None else a
        )
        self.data["org"] = self.data["url"].map(lambda a: a.split("/")[3])
        self.data["repo"] = self.data["url"].map(lambda a: a.split("/")[4])
        self.data["thumbsup"] = self.data["reactions"].map(lambda a: a["totalCount"])
        self.data.drop(columns="reactions", inplace=True)



import pandas as pd
from pathlib import Path

DEFAULT_PATH_CACHE = Path("~/data_github_activity").expanduser()


def _cache_data(query_data, path_cache):
    if path_cache is True:
        path_cache = DEFAULT_PATH_CACHE
    path_cache = Path(path_cache)
    if not path_cache.exists():
        print(f"Creating a new cache at {path_cache}")
        path_cache.mkdir()

    for (org, repo), idata in query_data.groupby(["org", "repo"]):
        path_repo_cache = path_cache.joinpath(org, repo)
        path_repo_cache.mkdir(parents=True, exist_ok=True)

        # First pull issues
        def _categorize_item(item):
            if f"{repo}/issues" in item:
                out = "issue"
            else:
                out = "pr"
            return out

        query_data["kind"] = query_data["url"].map(_categorize_item)

        # PRs cache
        data_prs = query_data.query("kind == 'pr'")
        path_pr_cache = path_repo_cache.joinpath("prs.csv")
        if path_pr_cache.exists():
            prs_cache = pd.read_csv(path_pr_cache)
            prs_cache = prs_cache.append(data_prs, sort=False).drop_duplicates(
                subset=["url"]
            )
        else:
            prs_cache = data_prs
        prs_cache.to_csv(path_pr_cache, index=False)

        # Issues cache
        data_issues = query_data.query("kind == 'issue'")
        path_issues_cache = path_repo_cache.joinpath("issues.csv")
        if path_issues_cache.exists():
            issues_cache = pd.read_csv(path_pr_cache)
            issues_cache = issues_cache.append(data_issues, sort=False).drop_duplicates(
                subset=["url"]
            )
        else:
            issues_cache = data_issues
        issues_cache.to_csv(path_issues_cache, index=False)

        # Comments collect
        data_comments = []
        for _, iitems in idata.iterrows():
            for icomment in iitems["comments"]["edges"]:
                this_comment = {}
                this_comment["author"] = icomment["node"]["author"]["login"]
                if not this_comment["author"]:
                    # This happens if the GitHub user has been deleted
                    # ref: https://github.com/jupyterhub/oauthenticator/pull/224#issuecomment-453211986
                    continue
                this_comment["created"] = icomment["node"]["createdAt"]
                this_comment["url"] = icomment["node"]["url"]
                data_comments.append(this_comment)
        data_comments = pd.DataFrame(data_comments).rename(
            columns={"created": "createdAt"}
        )

        # Comments cache
        path_comments_cache = path_repo_cache.joinpath("comments.csv")
        if path_comments_cache.exists():
            comments_cache = pd.read_csv(path_comments_cache)
            comments_cache = comments_cache.append(
                data_comments, sort=False
            ).drop_duplicates(subset=["url"])
        else:
            comments_cache = data_comments
        comments_cache.to_csv(path_comments_cache, index=False)


ALLOWED_KINDS = ["issues", "comments", "prs"]


def load_from_cache(target, kind, path_cache=None):
    # Checks for correctness and existence of cache
    if path_cache is None:
        path_cache = DEFAULT_PATH_CACHE
    if kind not in ALLOWED_KINDS:
        raise ValueError(f"Kind must be one of {ALLOWED_KINDS}, not {kind}")
    path_cache = Path(path_cache)

    # Resolve org / repo
    parts = target.split("/")
    if len(parts) == 1:
        org = parts[0]
        repo = None
    elif len(parts) == 2:
        org, repo = parts
    else:
        raise ValueError(f"Target must be of form org/repo, got: {target}")

    path_cache_org = path_cache.joinpath(org)
    if not path_cache_org.exists():
        raise ValueError(f"Could not find cache for org: {org}")
    if repo is None:
        repos = [
            ii
            for ii in path_cache_org.glob("*")
            if ii.is_dir() and not ii.name.startswith(".")
        ]
    else:
        repos = [path_cache_org.joinpath(repo)]

    # Now loop through the repos and grab data from the cache
    data = []
    for irepo in repos:
        path_cache_repo = path_cache_org.joinpath(irepo)
        if not path_cache_repo.exists():
            raise ValueError(f"Could not find cache for org/repo: {org}/{repo}")
        path_csv = path_cache_repo.joinpath(f"{kind}.csv")
        data.append(pd.read_csv(path_csv))
    data = pd.concat(data)
    return data


def get_cache_stats(path_cache=None):
    if path_cache is None:
        path_cache = DEFAULT_PATH_CACHE

    out_data = []
    for org in path_cache.glob("*"):
        for repo in org.glob("*"):
            for ipath in repo.glob("*"):
                kind = ipath.with_suffix("").name
                idata = pd.read_csv(ipath)
                mindate = idata["createdAt"].min()
                maxdate = idata["createdAt"].max()
                nrecords = idata.shape[0]
                out_data.append(
                    {
                        "org": org.name,
                        "repo": repo.name,
                        "kind": kind,
                        "start": mindate,
                        "end": maxdate,
                        "nrecords": nrecords,
                    }
                )
    return pd.DataFrame(out_data)



from setuptools import setup, find_packages
import os
import os.path as op
from glob import glob
from pathlib import Path

init = Path().joinpath("github_activity", "__init__.py")
for line in init.read_text().split("\n"):
    if line.startswith("__version__ ="):
        version = line.split(" = ")[-1].strip('"')
        break

# Source dependencies from requirements.txt file.
with open("requirements.txt", "r") as f:
    lines = f.readlines()
    install_packages = [line.strip() for line in lines]

setup(
    name="github_activity",
    version=version,
    include_package_data=True,
    python_requires=">=3.4",
    author="Chris Holdgraf",
    author_email="choldgraf@berkeley.edu",
    url="https://jupyter.org/",
    # this should be a whitespace separated string of keywords, not a list
    keywords="development changelog",
    description="Grab recent issue/PR activity from a GitHub repository and render it as markdown.",
    long_description=open("./README.md", "r").read(),
    long_description_content_type="text/markdown",
    license="BSD",
    packages=find_packages(),
    use_package_data=True,
    entry_points={"console_scripts": ["github-activity = github_activity.cli:main",]},
    install_requires=install_packages,
    extras_require={
        "testing": ["pytest", "pytest-regressions"],
        "sphinx": ["sphinx", "myst_parser", "sphinx_book_theme"],
    },
)
